#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <assert.h>
#include <getopt.h>
#include <errno.h>

#include <libfdt.h>

/* #define DEBUG */

static struct option long_options[] = {
	{"domain", 1, 0, 'd'},
	{"blob", 1, 0, 'b'},
	{"output", 1, 0, 'o'},
	{"help", 0, 0, 'h'},
	{0, 0, 0, 0}
};

static const char *usage =
"xen-cfg --domain <domain> --blob <fdt blob> --output <output file> --help\n"
"Usage options:\n"
"-h, --help             : Print this help\n";

static void print_usage(int exit_code)
{
	printf("%s", usage);
	exit(exit_code);
}

#define MAX_PATH 256
static char dom_kernel[MAX_PATH];
static int dom_memory;
static char dom_name[MAX_PATH];
static uint32_t dom_cpumask;

/*
 *    FDT Parsing
 */
static void *fdt;
enum {
	PROP_STR,
	PROP_INT
};

/* Helper functions to parse FDT tree */
int get_fdt_prop(const char* node_path, const char *prop, void *prop_buf, int type)
{
        int i;
        char * src, *dst;
	const void *pbuf;
	int prop_len;

        /* Get the Node */
        int node = fdt_path_offset(fdt, node_path);
        if (node < 0) {
                printf("%s: Failed to parse node path %s \n",
                           fdt_strerror(node), node_path);
                return -1;
        }
	/* Get the Prop */
        pbuf = (char *) fdt_getprop(fdt, node, prop, &prop_len);
        if (pbuf == NULL) {
                printf("%s: Failed to parse %s \n", fdt_strerror(prop_len), prop);
                return -1;
        }
        if (type == PROP_STR) {
                memcpy((void *) prop_buf, (void * )pbuf, prop_len);

        } else { /* Copy in reverse to account for host LE order */
                src = (char *) pbuf;
                dst = (char *) prop_buf;
                for (i = 0; i < prop_len; ++i) {
                        dst[i] = src[prop_len - i - 1];
                }
        }

#ifdef DEBUG
        prop_len /= 4;
        //printf("FDT: parsed %s.%s with len %d val: ", node_path, prop, prop_len);  
        if (type == PROP_INT) {
                for (i = 0; i < prop_len; ++i) {
                        printf("%d ", *((unsigned int *) prop_buf + i));
                }
        } else {
                printf("%s ", (char *) prop_buf);
        }
        printf("\n");
#endif

        return 0;
}

static void setup_dom_defaults(void)
{
	snprintf(dom_kernel, MAX_PATH, "%s", "/tmp/xen/netos.elf");

	/* 32MB Memory */
	dom_memory = 32;

	snprintf(dom_name, MAX_PATH, "%s", "NetOS");

	dom_cpumask = 0xf;
}

static void parse_blob(char *blob, int domain)
{
	struct stat st;
	int fd = -1;
	int fdt_err = 0;
	char dom_str[128];

	fd = open(blob, O_RDONLY);
	if (fd < 0) {
		perror("Failed to open FDT blob");
		exit(1);
	}

	if (fstat(fd, (struct stat *)&st) < 0) {
		perror(" Failed to read file status ");
		exit(1);
	}

	fdt = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (fdt ==  MAP_FAILED) {
		perror(" Failed to mmap cfg file ");
		exit(1);
	}

	fdt_err = fdt_check_header(fdt);
	if (fdt_err < 0) {
		printf("FDT_ERROR: Invalid FDT Detected: %s\n", fdt_strerror(fdt_err));
		exit(1);
	}
	printf("%s FDT blob header validated\n", __func__);

	snprintf(dom_str, 128, "/doms/dom@%d/kernel", domain);
	if (get_fdt_prop(dom_str, "location", dom_kernel, PROP_STR) != 0) {
		printf("Failed to parse kernel location! Aborting !\n");
		exit(1);
	}
	printf("dom_kernel = \"%s\"\n", dom_kernel);

	snprintf(dom_str, 128, "/doms/dom@%d/kernel", domain);
	if (get_fdt_prop(dom_str, "size", &dom_memory, PROP_INT) != 0) {
		printf("Failed to parse size! Aborting !\n");
		exit(1);
	}
	printf("dom_memory = 0x%08x\n", dom_memory);

	snprintf(dom_str, 128, "/doms/dom@%d", domain);
	if (get_fdt_prop(dom_str, "os", dom_name, PROP_STR) != 0) {
		printf("Failed to parse kernel location! Aborting !\n");
		exit(1);
	}
	printf("dom_name = \"%s\"\n", dom_name);

	snprintf(dom_str, 128, "/doms/dom@%d/cpu", domain);
	if (get_fdt_prop(dom_str, "onlinemask", &dom_cpumask, PROP_INT) != 0) {
		printf("Failed to parse onlinemask! Aborting !\n");
		exit(1);
	}
	printf("dom_cpumask = 0x%08x\n", dom_cpumask);
}

static void gen_cfg_file(int domain, char *filename)
{
	FILE *fp = 0;
	int num_cpus = 0;
	int i = 0;

	fp = fopen(filename, "w");
	if (!fp) {
		fprintf(stderr, "Unable to open filename: %s\n", strerror(errno));
		exit(-1);
	}

	fprintf(fp, "#  -*- mode: python; -*-\n\n");
	fprintf(fp, "kernel = \"%s\"\n\n", dom_kernel);
	fprintf(fp, "memory = %d\n\n", dom_memory >> 20);
	fprintf(fp, "name = \"%s@%d\"\n\n", dom_name, domain);

	fprintf(fp, "cpus = [");
	for (i = 0; i < 32; ++i) {
		if (dom_cpumask & (1 << i)) {

			num_cpus++;

			if (num_cpus != 1)
				fprintf(fp, ", ");
			fprintf(fp, "\"%d\"", (1 << i));
		}
	}
	fprintf(fp, "]\n\n");

	fprintf(fp, "vcpus = %d\n\n", num_cpus);

	fclose(fp);
	printf("Generated %s\n", filename);
}

int main(int argc, char *argv[])
{
	char c;
	int opt_idx;
	int domain = 1;
	char *blob = "sysconfig-xen.dtb";
	char *output = "sysconfig-xen-dom.cfg";

	while((c = getopt_long(argc, argv, "hb:d:o:", long_options, &opt_idx)) >= 0) {
		switch(c) {
		case 'd':
			domain = (int)strtoul(optarg, 0, 10);
			if( domain < 0) {
				printf("negative domain id (%d) specified?\n", domain);
				exit(-1);
			}
			break;
		case 'b':
			blob = strdup(optarg);
			break;
		case 'o':
			output = strdup(optarg);
			break;
		case 'h':
			print_usage(0);
			break;
		default:
			break;
		}
	}

	printf("domain = %d, blob = \"%s\", output = \"%s\"\n", domain, blob, output);

	setup_dom_defaults();

	parse_blob(blob, domain);

	gen_cfg_file(domain, output);

	return 0;
}
